/*
 * Kuklomenos
 * Copyright (C) 2008 Martin Bays <mbays@sdf.lonestar.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 */

#include <config.h>

#ifdef HAVE_LIBSDL_MIXER
#define MUSIC 1
#endif

#include <cstdlib>
#include <cstdio>
#include <ctime>
#include <stack>
#include <getopt.h>
#include <SDL/SDL.h>
#include <SDL_gfxPrimitivesDirty.h>

#ifdef MUSIC
#include <SDL/SDL_mixer.h>
#endif

#include "state.h"
#include "geom.h"
#include "clock.h"
#include "data.h"
#include "settings.h"
#include "conffile.h"
#include "random.h"
#include "menu.h"

#ifndef __APPLE__
# include "config.h"
#endif


#ifdef MUSIC
Mix_Music *music = NULL;
void checkMusic();
int musicStart();
void musicStop();
#endif

SDL_Surface* screen = NULL;
std::stack<Menu*> menuStack;
enum EventsReturn
{
    ER_NONE,
    ER_MISC,
    ER_QUIT,
    ER_SURRENDER,
    ER_RESTART,
    ER_SCREENSHOT,
    ER_MENU
};

EventsReturn process_events(GameState* gameState, GameClock& gameClock)
{
    SDL_Event event;
    bool wasEvent = false;
    while (SDL_PollEvent(&event))
    {
	wasEvent = true;
	switch (event.type)
	{
	    case SDL_QUIT:
		return ER_QUIT;
	    case SDL_KEYDOWN:
		{
		    SDLKey sym = event.key.keysym.sym;
		    switch (sym)
		    {
			case SDLK_q:
			    return ER_QUIT;
			case SDLK_p:
			    gameClock.paused = !gameClock.paused;
			    break;
			case SDLK_z:
			    settings.zoomEnabled = !settings.zoomEnabled;
			    break;
			case SDLK_r:
			    settings.rotatingView = !settings.rotatingView;
			    break;
			case SDLK_a:
			    if (!event.key.keysym.mod & KMOD_SHIFT)
				// 'a': increase anti-aliasing level
				settings.useAA =
				    (settings.useAA == AA_NO) ? AA_YES :
				    AA_FORCE;
			    else
				// 'A': decrease anti-aliasing level
				settings.useAA =
				    (settings.useAA == AA_FORCE) ? AA_YES :
				    AA_NO;
			    break;
			case SDLK_g:
			    settings.showGrid = !settings.showGrid;
			    break;
			case SDLK_LEFTBRACKET:
			    settings.fps = std::max(1, settings.fps-1);
			    break;
			case SDLK_RIGHTBRACKET:
			    settings.fps = std::min(100, settings.fps+1);
			    break;
			case SDLK_MINUS:
			case SDLK_KP_MINUS:
			    if (settings.debug || gameState->end)
			    {
				gameClock.rate = 4*gameClock.rate/5;
				if (950 <= gameClock.rate && gameClock.rate <= 1050)
				    gameClock.rate = 1000;
			    }
			    break;
			case SDLK_EQUALS:
			    if (!event.key.keysym.mod & KMOD_SHIFT)
			    {
				gameClock.rate = 1000;
				break;
			    }
			    // fallthrough
			case SDLK_PLUS:
			case SDLK_KP_PLUS:
			    if (settings.debug || gameState->end)
			    {
				gameClock.rate = std::max(5*gameClock.rate/4,
					gameClock.rate+1);
				if (950 <= gameClock.rate && gameClock.rate <= 1050)
				    gameClock.rate = 1000;
			    }
			    break;
			case SDLK_i:
			    if (settings.debug)
				settings.invuln = !settings.invuln;
			    break;
#ifdef MUSIC
			case SDLK_m:
			    settings.music = !settings.music;
			    checkMusic();
			    break;
#endif
			case SDLK_SPACE:
			case SDLK_RETURN:
			    if (gameState->end)
				return ER_RESTART;
			    break;

			case SDLK_F12:
			    return ER_SCREENSHOT;
			    break;

			case SDLK_ESCAPE:
			    if (menuStack.empty())
				menuStack.push(&topMenu);
			    else
				menuStack.pop();
			    return ER_MISC;
			    break;
			default: ;
		    }

		    if (!menuStack.empty())
		    {
			int dir = (sym == SDLK_h || sym == SDLK_h || sym == SDLK_LEFT) ? 0 :
			    (sym == SDLK_n || sym == SDLK_k || sym == SDLK_UP) ? 1 :
			    (sym == SDLK_s || sym == SDLK_l || sym == SDLK_RIGHT) ? 2 :
			    (sym == SDLK_t || sym == SDLK_j || sym == SDLK_DOWN) ? 3 : -1;

			if (dir >= 0)
			{
			    Menu* submenu = menuStack.top()->menus[dir];
			    MenuLeaf* leaf = menuStack.top()->leaves[dir];

			    if (submenu)
			    {
				menuStack.push(submenu);
			    }
			    else if (leaf)
			    {
				LeafReturn lr = leaf->act();
				switch (lr)
				{
				    case LR_QUIT:
					return ER_QUIT;
				    case LR_SURRENDER:
					while (!menuStack.empty())
						menuStack.pop();
					return ER_SURRENDER;
				    case LR_EXITMENU:
					menuStack.pop();
					break;
				    case LR_SAVESETTINGS:
					config.importSettings(settings);
					config.write();
					menuStack.pop();
					break;
				    default: ;
				}
#ifdef MUSIC
				checkMusic();
#endif
			    }
			    return ER_MISC;
			}
		    }
		    break;
		}
	    case SDL_VIDEORESIZE:
		screenGeom = ScreenGeom(event.resize.w, event.resize.h);
		screen = SDL_SetVideoMode(event.resize.w, event.resize.h,
			screen->format->BitsPerPixel, screen->flags);
		break;
	    default:
		return ER_NONE;
	}
    }
    return wasEvent ? ER_MISC : ER_NONE;
}

void drawInfo(SDL_Surface* surface, GameState* gameState, GameClock& gameClock,
	float observedFPS)
{
    char fpsStr[5+10+9];
    snprintf(fpsStr, 5+10+9, "fps: %.1f/%d%s", observedFPS, settings.fps,
		gameClock.paused ? " [Paused]" : "");

    char ratingStr[8+20+7];
    snprintf(ratingStr, 8+20+7, "rating: %s (%.1f)",
	    ratingString((int)(gameState->rating)), gameState->rating);

    char rateStr[6+5+10];
    snprintf(rateStr, 6+5+10, "speed: %d.%d%s", gameClock.rate/1000,
	    (gameClock.rate%1000)/100, settings.debug ? " [*DEBUG*]" : "");

    // if not enough room, try short version; if still too long don't display:
    if ((int)strlen(fpsStr) > screenGeom.infoMaxLength)
	snprintf(fpsStr, 5+10+9, "F: %.1f%s", observedFPS,
		gameClock.paused ? " [P]" : "");
    if ((int)strlen(fpsStr) > screenGeom.infoMaxLength)
	*fpsStr = '\0';

    if ((int)strlen(ratingStr) > screenGeom.infoMaxLength)
	snprintf(ratingStr, 8+20+7, "R: %.1f", gameState->rating);
    if ((int)strlen(ratingStr) > screenGeom.infoMaxLength)
	*ratingStr = '\0';

    if ((int)strlen(rateStr) > screenGeom.infoMaxLength)
	snprintf(rateStr, 6+5+9, "S: %d.%d%s", gameClock.rate/1000,
		(gameClock.rate%1000)/100, settings.debug ? " [D]" : "");
    if ((int)strlen(rateStr) > screenGeom.infoMaxLength)
	*rateStr = '\0';

    gfxPrimitivesSetFont(fontSmall,7,13);

    int line = 0;
    if (settings.showFPS)
	stringColor(surface,
		screenGeom.info.x, screenGeom.info.y+15*line++,
		fpsStr, 0xffffffff);


    if (screenGeom.infoMaxLines > line)
    {
	stringColor(surface,
		screenGeom.info.x, screenGeom.info.y+15*line++,
		ratingStr, 0xffffffff);
    }

    if ( (settings.debug || gameClock.rate != 1000) &&
	    screenGeom.infoMaxLines > line)
	stringColor(surface,
		screenGeom.info.x, screenGeom.info.y+15*line++,
		rateStr, 0xffffffff);
}

class OverlayText
{
    public:
	static const int text1Len = 60;
	static const int text2Len = 80;
	char text1[text1Len];
	char text2[text2Len];
	int colour;

	OverlayText();
	void blank();
	void draw(SDL_Surface* surface);
};

OverlayText::OverlayText() :
    colour(0)
{
    blank();
}
void OverlayText::blank()
{
    text1[0] = '\0';
    text2[0] = '\0';
}

void OverlayText::draw(SDL_Surface* surface)
{
    int x1 = screenGeom.centre.x - strlen(text1)*10/2;
    int y1 = screenGeom.centre.y - 30;
    int x2 = screenGeom.centre.x - strlen(text2)*10/2;
    int y2 = screenGeom.centre.y + 30;

    if (x1 > 0 && x2 > 0)
	gfxPrimitivesSetFont(fontBig, 10, 20);
    else
    {
	// not enough room for the big font...
	gfxPrimitivesSetFont(fontSmall, 7, 13);
	x1 = screenGeom.centre.x - strlen(text1)*7/2;
	y1 = screenGeom.centre.y - 20;
	x2 = screenGeom.centre.x - strlen(text2)*7/2;
	y2 = screenGeom.centre.y + 20;
    }

    stringColor(surface, x1, y1,
	    text1, colour);
    stringColor(surface, x2, y2,
	    text2, colour);
}

void drawMenu(SDL_Surface* surface, const Menu& menu)
{
    static const Uint32 colour = 0xffffffd0;

    filledCircleColor(surface, screenGeom.centre.x, screenGeom.centre.y+7, 15,
	    colour);
    aacircleColor(surface, screenGeom.centre.x, screenGeom.centre.y+7, 15,
	    colour);

    gfxPrimitivesSetFont(fontBig, 10, 20);
    for (int dir = 0; dir < 4; dir++)
    {
	Menu* submenu = menu.menus[dir];
	MenuLeaf* leaf = menu.leaves[dir];
	std::string text;

	if (submenu)
	    text = submenu->title;
	else if (leaf)
	    text = leaf->name();

	if (!text.empty())
	{
	    int x=0, y=0;
	    switch (dir)
	    {
		case 0:
		    x = screenGeom.centre.x - 50 - text.length()*10;
		    y = screenGeom.centre.y;
		    break;
		case 1:
		    x = screenGeom.centre.x - text.length()*10/2;
		    y = screenGeom.centre.y - 50;
		    break;
		case 2:
		    x = screenGeom.centre.x + 50;
		    y = screenGeom.centre.y;
		    break;
		case 3:
		    x = screenGeom.centre.x - text.length()*10/2;
		    y = screenGeom.centre.y + 50;
		    break;
		default:;
	    }
	    stringColor(surface, x, y, text.c_str(), colour);
	}
    }
}

const int MIN_INPUT_STEP = 30;
const int MIN_GAME_STEP = 30;

void initialize_system()
{
    /* Initialize SDL */
    if ( SDL_Init(SDL_INIT_EVERYTHING) < 0 ) {
	fprintf(stderr,
		"Couldn't initialize SDL: %s\n", SDL_GetError());
	exit(1);
    }
    atexit(SDL_Quit);			/* Clean up on exit */

    SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

    // set random seed
    srand(time(NULL));
}

#ifndef __APPLE__
void load_settings(int argc, char** argv)
{
    config.read();
    config.exportSettings(settings);

    while (1)
    {
	int c;
	static const struct option long_options[] =
	{
	    {"width", 1, 0, 'W'},
	    {"height", 1, 0, 'H'},
	    {"bpp", 1, 0, 'b'},
	    {"fps", 1, 0, 'f'},
	    {"rating", 1, 0, 'r'},
	    {"gamma", 1, 0, 'g' << 8},
	    {"noantialias", 0, 0, 'A'},
	    {"nogrid", 0, 0, 'G'},
	    {"nozoom", 0, 0, 'Z'},
	    {"norotate", 0, 0, 'R'},
	    {"antialias", 0, 0, 'a'},
	    {"grid", 0, 0, 'g'},
	    {"zoom", 0, 0, 'z'},
	    {"rotate", 0, 0, '@'},
	    {"fullscreen", 0, 0, 'F'},
	    {"hwsurface", 0, 0, 's'},
	    {"hwpalette", 0, 0, 'P'},
	    {"noresizable", 0, 0, 'S'},
	    {"music", 0, 0, 'm'},
	    {"debug", 0, 0, 'd'},
	    {"invulnerable", 0, 0, 'i'},
	    {"speed", 1, 0, 'p'},
	    {"version", 0, 0, 'V'},
	    {"help", 0, 0, 'h'},
	    {0,0,0,0}
	};
	c = getopt_long(argc, argv, "W:H:b:f:r:I:p:AGZRagzFsPSmdiVh", long_options, NULL);
	if (c == -1)
	    break;
	switch (c)
	{
	    case 'W':
		settings.width = atoi(optarg);
		break;
	    case 'H':
		settings.height = atoi(optarg);
		break;
	    case 'b':
		settings.bpp = atoi(optarg);
		break;
	    case 'f':
		settings.fps = atoi(optarg);
		break;
	    case 'r':
		if (1.0 <= atof(optarg))
		    settings.requestedRating = atof(optarg);
		else
		{
		    printf("bad rating\n");
		    exit(1);
		}
		break;
	    case 'g'<<8:
		antialiasGamma = atoi(optarg);
		break;
	    case 'F':
		settings.fullscreen = true;
		break;
	    case 'S':
		settings.videoFlags &= ~SDL_RESIZABLE;
		break;
	    case 'P':
		settings.videoFlags |= SDL_HWPALETTE;
		break;
	    case 's':
		settings.videoFlags |= SDL_HWSURFACE;
		break;
	    case 'Z':
		settings.zoomEnabled = false;
		break;
	    case 'R':
		settings.rotatingView = false;
		break;
	    case 'A':
		settings.useAA =
		    (settings.useAA == AA_FORCE) ? AA_YES :
		    AA_NO;
		break;
	    case 'G':
		settings.showGrid = false;
		break;
	    case 'z':
		settings.zoomEnabled = true;
		break;
	    case '@':
		settings.rotatingView = true;
		break;
	    case 'a':
		settings.useAA =
		    (settings.useAA == AA_NO) ? AA_YES :
		    AA_FORCE;
		break;
	    case 'g':
		settings.showGrid = true;
		break;
	    case 'm':
		settings.music = true;
		break;
	    case 'd':
		settings.debug = true;
		break;
	    case 'i':
		settings.invuln = true;
		settings.debug = true;
		break;
	    case 'p':
		settings.clockRate = (int)(1000*atof(optarg));
		settings.debug = true;
		break;
	    case 'V':
		printf("%s\n", PACKAGE_STRING);
		exit(0);
	    case 'h':
		printf("Options:\n\t"
			"-W --width WIDTH\n\t-H --height HEIGHT\n\t-b --bpp BITS\n\t-f --fps FPS\n\t"
			"-F --fullscreen\n\t-S --noresizable\n\t-P --hwpalette\n\t-s --hwsurface\n\t"
			"-Z,-z --[no]zoom\n\t-R --[no]rotate\n\t-G,-g --[no]grid\n\t-A,-a --[no]antialias\n\t"
			"-r --rating RATING\t\t1: harmless... 4: average... 9: elite\n\t"
			"--gamma GAMMA\n\t"
#ifdef MUSIC
			"-m --music\n\t"
#endif
			"-d --debug\n\t-i --invulnerable\n\t-p --speed RATE\n\t"
			"-V --version\n\t-h --help\n");
		exit(0);
		break;

	    case '?':
		exit(1);
		break;

	    default:
		printf("getopt returned unexpected character %c\n", c);
	}
    }
}
#endif

void initialize_video()
{
    if (settings.fullscreen)
	settings.videoFlags |= SDL_FULLSCREEN;

    if (settings.width == 0 || settings.height == 0)
    {
	SDL_Rect** modes = SDL_ListModes(NULL, settings.videoFlags |
		(settings.fullscreen ? SDL_FULLSCREEN : 0));
	if (modes == NULL)
	{
	    fprintf(stderr,
		    "SDL reports no modes available\n");
	    exit(1);
	}
	if (modes == (SDL_Rect**)(-1))
	{
	    // "All modes available"
	    settings.width = 1024;
	    settings.height = 768;
	}
	else
	{
	    // use the first (i.e. biggest) mode which is no bigger than
	    // 1024x768
	    SDL_Rect* bestMode;
	    do
	    {
		bestMode = *modes;
		if (bestMode->w <= 1024 && bestMode->h <= 768)
		    break;
	    } while (++modes);
	    settings.width = bestMode->w;
	    settings.height = bestMode->h;
	}
    }

    /* Initialize the display */
    screen = SDL_SetVideoMode(settings.width, settings.height, settings.bpp, settings.videoFlags);
    if ( screen == NULL ) {
	fprintf(stderr, "Couldn't set %dx%dx%d video mode: %s\n",
		settings.width, settings.height, settings.bpp, SDL_GetError());
	exit(1);
    }

    /* Show some info */
    printf("Set %dx%dx%d mode\n",
	    screen->w, screen->h, screen->format->BitsPerPixel);
    printf("Video surface located in %s memory.\n",
	    (screen->flags&SDL_HWSURFACE) ? "video" : "system");

    /* Check for double buffering */
    if ( screen->flags & SDL_DOUBLEBUF ) {
	printf("Double-buffering enabled - good!\n");
    }

    SDL_WM_SetCaption( "Kuklomenos", NULL );
    SDL_ShowCursor(SDL_DISABLE);

    if ( !initFont() )
    {
	fprintf(stderr, "Failed to load font data file\n");
	exit(1);
    }

    screenGeom = ScreenGeom(settings.width, settings.height);
}

#ifdef MUSIC

void checkMusic()
{
    if (settings.music && !music)
	musicStart();
    else if (music && !settings.music)
	musicStop();
}

int musicStart()
{
    if ( Mix_OpenAudio(44100, AUDIO_S16, 1, 4096) < 0)
    {
	fprintf(stderr, "Mix_OpenAudio: %s\n", Mix_GetError());
	return -1;
    }

    const string path = findDataPath("amiga_machines.mod");
    if (path == "")
    {
	fprintf(stderr, "musicStart: can't find music file\n");
	return -1;
    }

    music = Mix_LoadMUS(findDataPath("amiga_machines.mod").c_str());

    if (!music)
    {
	fprintf(stderr, "Mix_LoadMUS: %s\n", Mix_GetError());
	return -1;
    }

    if ( Mix_PlayMusic(music, -1) < 0 )
    {
	fprintf(stderr, "Mix_PlayMusic: %s\n", Mix_GetError());
	return -1;
    }

    return 0;
}

void musicStop()
{
    if (music)
    {
	Mix_HaltMusic();
	Mix_FreeMusic(music);
	music = NULL;
    }
}
#endif

void run_game()
{
    float initialRating;
    if (settings.requestedRating > 0)
	initialRating = settings.requestedRating;
    else
    {
	initialRating = config.rating;
	if (!settings.debug)
	    config.shouldUpdateRating = true;
    }

    GameState* gameState = new GameState(initialRating);

    GameClock gameClock(settings.clockRate);

    OverlayText overlayText;
	
    // main loop
    Uint32 lastStateUpdate = 0, beforeDelay = 0;
    Uint32 ticksBefore, ticksAfter;
    Uint32 loopTicks = SDL_GetTicks();
    int timeTillNextFrame = 0;
    int delayTime, actualDelayed, updateTime;
    float avFrameTime = 1000/settings.fps;
    int fpsRegulatorTenthTicks = 0;
    const int avFrames = 10; // number of frames to average over
    EventsReturn eventsReturn = ER_NONE;
    bool wantScreenshot = false;
    bool quit = false;
    bool ended = false;
    bool forceFrame = false;
    while ( !quit ) {
	forceFrame = false;
	while (timeTillNextFrame > 0)
	{
	    delayTime = std::min(timeTillNextFrame, MIN_INPUT_STEP);
	    beforeDelay = SDL_GetTicks();
	    SDL_Delay(delayTime);
	    actualDelayed = SDL_GetTicks() - beforeDelay;
	    timeTillNextFrame -= actualDelayed;

	    eventsReturn = process_events(gameState, gameClock);

	    switch (eventsReturn)
	    {
		case ER_RESTART:
		    {
			GameState* newGameState = new GameState(gameState->rating);
			delete gameState;
			gameState = newGameState;
			gameClock = GameClock(settings.clockRate);
		    }
		    ended = false;
		    overlayText.blank();
		    break;
		case ER_QUIT:
		    quit = true;
		    break;
		case ER_SURRENDER:
		    if (!ended)
			gameState->end = END_DEAD;
		case ER_SCREENSHOT:
		    wantScreenshot = true;
		    break;
		case ER_MISC:
		    forceFrame = true;
		    break;
		default: ;
	    }

	    updateTime = SDL_GetTicks() - lastStateUpdate;
	    lastStateUpdate = SDL_GetTicks();
	    if (!gameClock.paused && (menuStack.empty() || ended))
	    {
		while (updateTime > 0)
		{
		    const int stepTime = std::min(updateTime, MIN_GAME_STEP);
		    gameState->update(gameClock.scale(stepTime),
			    !menuStack.empty());
		    gameClock.update(stepTime);
		    updateTime -= stepTime;
		}
	    }
	}

	ticksBefore = SDL_GetTicks();
	if (!gameClock.paused || forceFrame)
	{
	    gameState->draw(screen);
	    drawInfo(screen, gameState, gameClock, 1000.0/avFrameTime);
	    overlayText.draw(screen);
	    if (!menuStack.empty())
		drawMenu(screen, *menuStack.top());
	    SDL_Flip(screen);

	    if (wantScreenshot)
	    {
		SDL_SaveBMP(screen, "screenshot.bmp");
		wantScreenshot = false;
	    }

	    // blank over what we've drawn:
	    blankDirty(screen);
	}
	ticksAfter = SDL_GetTicks();

	const int renderingTicks = ticksAfter - ticksBefore;

	// Add in a manual tweak to the delay, to ensure we approximate the
	// requested fps (where possible):
	if (10000/avFrameTime < settings.fps * 10 &&
		1000/settings.fps - renderingTicks + fpsRegulatorTenthTicks/10 > 1)
	    fpsRegulatorTenthTicks -= 1;
	else if (10000/avFrameTime > settings.fps * 10)
	    fpsRegulatorTenthTicks += 1;

	timeTillNextFrame =
	    std::max(1, (1000/settings.fps) - renderingTicks +
		    fpsRegulatorTenthTicks/10 + (rani(10) < fpsRegulatorTenthTicks%10));

	if (!ended && gameState->end)
	{
	    ended = true;
	    switch (gameState->end)
	    {
		case END_DEAD:
		case END_EXTRACTED:
		    snprintf(overlayText.text1, overlayText.text1Len, "Failure.");
		    overlayText.colour = 0xff4444ff;
		    break;
		case END_WIN:
		    snprintf(overlayText.text1, overlayText.text1Len, "Victory!");
		    overlayText.colour = 0x88ff00ff;
		    break;
		default: ;
	    }

	    // give a hint if appropriate
	    strncpy(overlayText.text2, gameState->getHint(),
		    overlayText.text2Len);

	    if (config.shouldUpdateRating)
	    {
		int old = (int)(gameState->rating);

		if (gameState->end == END_DEAD ||
			gameState->end == END_EXTRACTED)
		{
		    gameState->rating -= 0.1;
		    if (gameState->rating < 1)
			gameState->rating = 1.0;

		}
		else if (gameState->end == END_WIN)
		{
		    gameState->rating += 0.4;
		    if (gameState->rating < 5)
			// extra boost, so clawing your way back up from the
			// depths you sink to while first learning the game isn't
			// too arduous:
			gameState->rating += 0.2*(5 - (int)gameState->rating);
		}

		// deal with float inaccuracy - sometimes, 0.6+0.4 is just
		// less that 1.0...
		if ((gameState->rating - (int)(gameState->rating)) > 0.95)
		    gameState->rating = (int)(gameState->rating) + 1;

		if ((int)(gameState->rating) != old)
		{
		    if (overlayText.text2[0] == '\0')
		    {
			snprintf(overlayText.text2, overlayText.text2Len, "New rating: \"%s\".",
				ratingString( (int)(gameState->rating) ));
		    }
		    else
		    {
			char buf[40];
			snprintf(buf, 40, "  New rating: \"%s\".",
				ratingString( (int)(gameState->rating) ));
			strncat(overlayText.text1, buf,
				overlayText.text1Len - strlen(overlayText.text1) - 1);
		    }
		}
	    }
	}

	const int loopTime = SDL_GetTicks() - loopTicks;
	avFrameTime += (loopTime-avFrameTime)/avFrames;
	loopTicks = SDL_GetTicks();
    }

    if (!ended && gameState->extracted > 0)
    {
	// game is forfeit - reduce rating
	gameState->rating -= 0.1;
	if (gameState->rating < 1)
	    gameState->rating = 1.0;
    }
    if (config.shouldUpdateRating)
    {
	config.rating = gameState->rating;
	config.write();
    }

    delete gameState;

#ifdef MUSIC
    musicStop();
#endif

    SDL_Quit();
}

#ifndef __APPLE__
int main(int argc, char** argv)
{
    load_settings(argc, argv);
#ifdef MUSIC
    if (settings.music)
	musicStart();
#endif
    initialize_system();
    initialize_video();
    run_game();

    return 0;
}
#endif /* __APPLE__ */
